#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "string.h"
#include "common.h"
#include "indexer.h"
#include "dbms.h"
#include "parseurl.h"
#include "spell.h"
#include "charset.h"
#include "log.h"
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif


#if (WIN32|WINNT)
#ifdef HAVE_PTHREAD
static CRITICAL_SECTION main_mutex;
#endif
#else
#include <unistd.h>
#ifdef HAVE_PTHREAD
static pthread_mutex_t main_mutex;
#endif
#endif


static int seconds=0;
static int nthreads=0;
static index_flags=0;
#ifdef HAVE_PTHREAD
static int thread_number=1;
#else
static int thread_number=0;
#endif

static char *config_name= CONFIGDIR "/indexer.conf";


int usage(){

	fprintf(stderr,
"\n\
indexer from UdmSearch v." VERSION "/%s\n\
http://mysearch.udm.net (C) 1998-2000, UdmSearch Developers Team.\n\
\n\
Usage: indexer [OPTIONS]  [configfile]\n\
\n\
Indexing options:\
"
#ifdef HAVE_SQL
"\n\
  -a            reindex all documents even if not expired (may be\n\
                limited using -t, -u, -s and -f options)\n\
  -m            reindex expired documents even if not modified (may\n\
                be limited using -t, -u, and -s options)\n\
  -e            index 'most expired' (oldest) documents first\n\
  -o            index documents with less depth (hops value) first\n\
  -n n          index only n documents and exit\n\
  -q            quick startup (do not add Server URLs)\n\
"
#endif
"\n\
  -i            insert new URLs (URLs to insert must be given using -u or -f)\n\
  -p n          sleep n seconds after each URL\n\
  -w            do not warn before clearing documents from database\n\
"
#ifdef HAVE_PTHREAD
"  -N n         run N threads\n\
"
#endif

#ifdef HAVE_SQL
"\n\
Subsection control options (may be combined):\n\
  -s status     limit indexer to documents matching status (HTTP Status code)\n\
  -t tag        limit indexer to documents matching tag\n\
  -u pattern    limit indexer to documents with URLs matching pattern\n\
                (supports SQL LIKE wildcard '%%')\n\
  -f filename   read URLs to be indexed/inserted/cleared from file (with -a\n\
                or -C option, supports SQL LIKE wildcard '%%'; has no effect\n\
                when combined with -m option)\n\
"
#else
"\n\
URL options:\n\
  -u URL        insert URL at startup\n\
  -f filename   read URLs to be inserted from file\n\
"
#endif
"\n\
Logging options:\n\
"
#ifdef LOG_PERROR
"  -l            do not log to stdout/stderr\n\
"
#endif
"  -v n          verbose level, 0-5\n\
\n\
Misc. options:\n\
"
#ifdef HAVE_SQL
"\n\
  -C            clear database and exit\n\
  -S            print statistics and exit\n\
  -I            print referers and exit\n\
"
#endif
"  -h,-?         print this help page and exit\n\
\n\
\n\
Mail bug reports and suggestions to <udmsearch@search.udm.net>.\n",DBVersion);

	return(0);
}



static int IndexerMain(INDEXER * Indexer){
int res=0;

	while(1){
		res=IndexNextURL(Indexer,index_flags);
		if(seconds){
			udm_log(Indexer->handle,UDM_LOG_DEBUG,"Sleeping %d second(s)",seconds);
			UDMSLEEP(seconds);
		}
		switch(res){
			case IND_OK:
				continue;				
			case IND_ERROR:
				THREADINFO(Indexer->handle,"Error",IndexerErrorMsg(Indexer));
				udm_log(Indexer->handle,UDM_LOG_ERROR,"Error: '%s'",IndexerErrorMsg(Indexer));
				return(1);
			case IND_NO_TARGET:
#ifdef HAVE_PTHREAD
			/* in multi-threaded environment we		*/
			/* should wait for a moment when every thread	*/
			/* has nothing to do				*/

				INC_ACC();

				LOCK_MAIN();
				UDMSLEEP(1);
				res=(GetActive()>=nthreads);
				UNLOCK_MAIN();

				DEC_ACC();
				if(res)return(0);
				continue;
#else
				return(0);
#endif
			default:
				return(0);
		}
	}
	return(0);
}


INDEXER * I[100];
#if  (WIN32|WINNT)
DWORD WINAPI thread_main(void *arg){
#else
void * thread_main(void *arg){
#endif
int mynum;

	LOCK_MAIN();
	mynum=thread_number++;
	I[mynum]=AllocIndexer(mynum);
	UNLOCK_MAIN();

	IndexerMain(I[mynum]);

	LOCK_MAIN();
	nthreads--;
	udm_log(I[mynum]->handle,INFO,"Done");
	FreeIndexer(I[mynum]);
	I[mynum]=0;
	UNLOCK_MAIN();
#if     (WIN32|WINNT)
	return(0);
#else
	return(NULL);
#endif
}


int main(int argc, char **argv) {

char *language=NULL;
int npages=-1;
int clear=0,stat=0,integrity=0;
int log2stderr=1,dump=0;
int add_servers=1;
extern char *optarg;
extern int optind;
int ch;
int maxthreads=1;

int warnings=1;

	while ((ch = getopt(argc, argv, "CSIaheolmdqiw?t:u:s:n:v:L:A:D:p:N:f:")) != -1){
		switch (ch) {
		case 'C': clear++;add_servers=0;break;	
		case 'S': stat++;add_servers=0;break;
		case 'I': integrity++;add_servers=0;break;
		case 'L': language=optarg;break;
		case 'q': add_servers=0;break;
		case 'l': log2stderr=0;break;
		case 'a': index_flags|=FLAG_MARK;break;
		case 'e': index_flags|=FLAG_EXP_FIRST;break;
		case 'o': index_flags|=FLAG_SORT_HOPS;break;
		case 'm': index_flags|=FLAG_REINDEX;break;
		case 'n': npages=atoi(optarg);break;
		case 'v': SetLogLevel(atoi(optarg));break;
		case 'p': seconds=atoi(optarg);break;
		case 'd': dump=1;break;
		case 't': AddTagLimit(atoi(optarg));break;
		case 's': AddStatusLimit(atoi(optarg));break;
		case 'u': 
			AddURLLimit(optarg);
			if(index_flags&FLAG_INSERT)
				AddHref(NULL,optarg,0,0,0);
			break;
		case 'N': maxthreads=atoi(optarg);break;
		case 'f': url_file_name=optarg;break;
		case 'i': index_flags|=FLAG_INSERT;break;
		case 'w': warnings=0;break;
		case '?':
		case 'h':
		default:  
			usage();
			return(1);
		}
	}

	argc -= optind;argv += optind;

	if(argc>1){
		usage();
		return(1);
	}
	if(argc==1)config_name=argv[0];

	InitCharset();

#ifdef HAVE_PTHREAD
	InitMutexes();
#if (WIN32|WINNT)
	InitializeCriticalSection(&main_mutex);
#else
	pthread_mutex_init(&main_mutex,NULL);
#endif
#endif

	if(LoadConfig(config_name,0,
		(add_servers?FLAG_ADD_SERV:0)+
		((!integrity&&!stat&&!clear)*FLAG_SPELL))){		
			fprintf(stderr,"%s\n",ConfErrMsg());
			exit(1);
	}

#if (WIN32|WINNT)
#else
	udm_openlog(log2stderr);
#endif

	/* Make sure URL file is readable */
	if(url_file_name) {
		FILE *url_file;
		if(!(url_file=fopen(url_file_name,"r"))){
			udm_log(0,UDM_LOG_ERROR,"Error: can't open url file '%s': %s",url_file_name, strerror(errno));
			exit(1);
		}
		fclose(url_file);
	}

	/* Make sure all URLs to be inserted are OK */
	if(index_flags&FLAG_INSERT) {
		/*
		if(uurl) {
			URL myurl;
			if(parseurl(&myurl,uurl)) {
				udm_log(0,UDM_LOG_ERROR,"Error: Invalid URL '%s'",uurl);
				exit(1);
			}
		}
		*/

		if(url_file_name) {
			if(IND_OK!=URLFile(NULL, URL_FILE_PARSE)){
				udm_log(0,UDM_LOG_ERROR,"Error: Invalid URL in '%s'",url_file_name);
				exit(1);
			}
		}
	}

	if(clear){
		INDEXER * Indexer;
		int clear_confirmed=0;
		if(warnings) {
			char str[5]="";
			printf("You are going to delete database '%s' content\n",DBName?DBName:"");
			printf("Are you sure?(YES/no)");
			if(fgets(str,sizeof(str),stdin)){
				if(strncmp(str,"YES",3))
					printf("Canceled\n");
				else
					clear_confirmed=1;
			}
		}
		else
			clear_confirmed=1;

		if(clear_confirmed) {
			Indexer=AllocIndexer(0);
			if(url_file_name) {
				if(IND_OK!=URLFile(Indexer,URL_FILE_CLEAR)){
					udm_log(Indexer->handle,UDM_LOG_ERROR,"Error: '%s'",IndexerErrorMsg(Indexer));
				}
			}
			else {
				if(IND_OK!=ClearDatabase(Indexer)){
					udm_log(Indexer->handle,UDM_LOG_ERROR,"Error: '%s'",IndexerErrorMsg(Indexer));
				}
			}
			FreeIndexer(Indexer);
		}else{
			printf("Canceled\n");
		}
	}else
	if(stat){
		INDEXER * Indexer;
		Indexer=AllocIndexer(0);
		if(IND_OK!=ShowStatistics(Indexer)){
			udm_log(Indexer->handle,UDM_LOG_ERROR,"Error: '%s'",IndexerErrorMsg(Indexer));
		}
		FreeIndexer(Indexer);
	}else
	if(integrity){
		INDEXER * Indexer;
		Indexer=AllocIndexer(0);
		if(IND_OK!=ShowReferers(Indexer)){
			udm_log(Indexer->handle,UDM_LOG_ERROR,"Error: '%s'",IndexerErrorMsg(Indexer));
		}
		FreeIndexer(Indexer);
	}else{
#if (WIN32|WINNT)
#else
		udm_log(0,INFO, "indexer from UdmSearch v." VERSION"/%s started with '%s'", DBVersion, config_name);
#endif
		SetMaxURLNumber(npages);

#ifdef HAVE_PTHREAD
		{
#if (WIN32|WINNT)
#else
			pthread_t *thread;
#endif
			int i;
			thread_number=1;
			for(i=0;i<maxthreads;i++){
#if (WIN32|WINNT)
#else
				thread=(pthread_t*)malloc(sizeof(pthread_t));
#endif
				LOCK_MAIN();
				nthreads++;
				UNLOCK_MAIN();
#if (WIN32|WINNT)
				CreateThread(NULL, 0, &thread_main, NULL, 0, NULL);
#else
				pthread_create(thread,NULL,&thread_main,NULL);
#endif
			}
		}
		
		while(1){
			LOCK_MAIN();
			if(!nthreads){
				UNLOCK_MAIN();
				break;
			}
			UNLOCK_MAIN();
			UDMSLEEP(1);
		}
		DestroyMutexes();

#if (WIN32|WINNT)
		DeleteCriticalSection(&main_mutex);
#else
		pthread_mutex_destroy(&main_mutex);
#endif
#else
		thread_main(NULL);
#endif
	}
	thread_number=0;

	return(0);
}
